/**************************************************************************
*   Copyright (C) 2005-2006 by Jinsoo Kang                                *
*   jsk@gravitybind.com                                                   *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
**************************************************************************/

#include "tx.cfg"
#include "tx_util.h"
#include "xmltable.h"

#include <iostream>

using namespace std;


Q3IntDict<XmlTableStyle> XmlTable::styleVector = Q3IntDict<XmlTableStyle>();
int XmlTable::minColumnWidth = 50;
int XmlTable::indentWidth = 30;
QColor XmlTable::borderColor(0,0,0);
QColor XmlTable::emptyBorderColor("lightgray");
QColor XmlTable::focusBorderColor(0,0,0);
bool XmlTable::displayBranches = true;
bool XmlTable::sortAttr = true;
QString XmlTable::outputEncoding;


XmlTable::XmlTable( int r, int c, QWidget * parent) : Q3Table( r, c, parent) {
	dashPattern << 1 << 6;
	max_cols = 0;
	max_rows = 0;
	applySettings();
	indent = "\t";
	xmlVersion = "1.0";
	setModified(FALSE);
	setCurrentCell(0, 0);
	setFocusStyle(Q3Table::FollowStyle);
	setPaletteBackgroundColor( style(XmlTableItem::EMPTY)->bgColor() );
	setPaletteForegroundColor( style(XmlTableItem::EMPTY)->fgColor() );

	setLeftMargin( fontMetrics().width("W") * QString::number(r).length() );

	connect(
		this, SIGNAL(pressed( int, int, int, const QPoint &)),
		this, SLOT(relayMousePress( int, int, int, const QPoint &))
		);
	connect(
		this, SIGNAL(currentChanged( int, int )),
		this, SLOT(endEditAtLastLocation())
		);

	tmpFileIndex = -1;
	resetUndo();
	resetRedo();
	QString undoFile;
	do {
		tmpFileIndex++;
		undoFile = tempFilePath("undo", tmpFileIndex, 0);
	} while (QFile::exists(undoFile));
	QFile firstFile(undoFile);
	if (!firstFile.open(QIODevice::WriteOnly)) return;
	QTextStream stream( &firstFile );
	stream << endl;
	firstFile.flush();
	firstFile.close();
}


XmlTable::~XmlTable() {
	QDir tdir = QDir::home();
	tdir.cd(".tx");
	QStringList ls = tdir.entryList( QString("??do.%1.*").arg(tmpFileIndex), QDir::Files );
	for ( QStringList::Iterator it = ls.begin(); it != ls.end(); ++it ) {
		QFile::remove(tdir.absPath() + tdir.separator() + (*it));
	}
}


void XmlTable::endEditAtLastLocation() {
	if (editingRow < 0 || editingColumn < 0) return;
	endEdit(editingRow, editingColumn, TRUE, FALSE);
	setEditingLocation();
}


void XmlTable::resetUndo() {
	undoIndex = 0;
	emit undoRecordSizeChanged(undoIndex);
}


void XmlTable::resetRedo() {
	redoIndex = 0;
	emit redoRecordSizeChanged(redoIndex);
}


QString XmlTable::tempFilePath(QString type, int i, int index) {
	QDir homeDir = QDir::home();
	homeDir.mkdir(".tx");
	QString tmpPath = QString(
		homeDir.absPath() + homeDir.separator() 
		+ ".tx" + homeDir.separator()
		+ "%1.%2.%3"
		).arg(type).arg(i).arg(index);
	return tmpPath;
}


QString XmlTable::newTempFilePath(QString type, int i, int &index) {
	QDir homeDir = QDir::home();
	homeDir.mkdir(".tx");
	QString tmpPath;
	do {
		index++;
		tmpPath = QString(
			homeDir.absPath() + homeDir.separator() 
			+ ".tx" + homeDir.separator()
			+ "%1.%2.%3"
			).arg(type).arg(i).arg(index);
	} while ( QFile::exists(tmpPath) );
	return tmpPath;
}


void XmlTable::addUndoRecord(int topRow, int leftCol, int bottomRow, int rightCol, bool fromRedo) {
	QFile undoFile(tempFilePath("undo", tmpFileIndex, ++undoIndex));
	if (!undoFile.open(QIODevice::WriteOnly)) return;
	if (!fromRedo) resetRedo();
	QTextStream stream( &undoFile );
	stream << endl 
		<< QString("<table rA='%1' cA='%2' rZ='%3' cZ='%4'>").arg(topRow).arg(leftCol).arg(bottomRow).arg(rightCol)
		<< endl;
	for (int r=topRow; r <= bottomRow; r++) {
		stream << "<tr>" << endl;
		for (int c = leftCol; c <= rightCol; c++) {
			XmlTableItem *currItem = xmlItem(r, c);
			if (currItem != NULL) {
				stream << QString("<td class='%1' depth='%2'>")
					.arg( style(currItem->itemType())->name() )
					.arg( currItem->depth()-1 );
			}
			else stream << "<td>";
			stream << encodeToXml(text(r, c));
			stream << "</td>" << endl;
			if (currItem != NULL && currItem->isSpanCell()) {
				c++;
				while (c <= rightCol) {
					stream << "<td class='empty' depth='0'/>" << endl;
					c++;
				}
			}
		}
		stream << "</tr>" << endl;
	}
	stream << QString("</table>") << endl;
	undoFile.flush();
	undoFile.close();

	if (undoIndex > 1) {
		QFile newUndoFile(tempFilePath("undo", tmpFileIndex, undoIndex));
		QFile oldUndoFile(tempFilePath("undo", tmpFileIndex, undoIndex-1));
		if (noDiff( &newUndoFile, &oldUndoFile )) undoIndex--;
		newUndoFile.close();
		oldUndoFile.close();
	}
	emit undoRecordSizeChanged(undoIndex);
}


void XmlTable::addRedoRecord(int topRow, int leftCol, int bottomRow, int rightCol) {
	QFile redoFile(tempFilePath("redo", tmpFileIndex, ++redoIndex));
	if (!redoFile.open(QIODevice::WriteOnly)) return;
	QTextStream stream( &redoFile );
	stream << endl 
		<< QString("<table rA='%1' cA='%2' rZ='%3' cZ='%4'>").arg(topRow).arg(leftCol).arg(bottomRow).arg(rightCol)
		<< endl;
	for (int r=topRow; r <= bottomRow; r++) {
		stream << "<tr>" << endl;
		for (int c = leftCol; c <= rightCol; c++) {
			XmlTableItem *currItem = xmlItem(r, c);
			if (currItem != NULL) {
				stream << QString("<td class='%1' depth='%2'>")
					.arg( style(currItem->itemType())->name() )
					.arg( currItem->depth()-1 );
			}
			else stream << "<td>";
			stream << encodeToXml(text(r, c));
			stream << "</td>" << endl;
			if (currItem != NULL && currItem->isSpanCell()) {
				c++;
				while (c <= rightCol) {
					stream << "<td class='empty' depth='0'/>" << endl;
					c++;
				}
			}
		}
		stream << "</tr>" << endl;
	}
	stream << QString("</table>") << endl;
	redoFile.flush();
	redoFile.close();

	if (redoIndex > 1) {
		QFile newRedoFile(tempFilePath("redo", tmpFileIndex, redoIndex));
		QFile oldRedoFile(tempFilePath("redo", tmpFileIndex, redoIndex-1));
		if (noDiff( &newRedoFile, &oldRedoFile )) redoIndex--;
		newRedoFile.close();
		oldRedoFile.close();
	}
	emit redoRecordSizeChanged(redoIndex);
}


bool XmlTable::noDiff(QFile *f1, QFile *f2) {
	QTextStream t1(f1), t2(f2);
	QString curr1, curr2;
	if (!f1->open(QIODevice::ReadOnly)) return FALSE;
	if (!f2->open(QIODevice::ReadOnly)) return FALSE;
	do {
		curr1 = t1.readLine();
		curr2 = t2.readLine();
		if (curr1 != curr2) {
			return FALSE;
		}
	}
	while (!curr1.isNull() && !curr2.isNull());
	if (!curr1.isNull() || !curr2.isNull()) {
		return FALSE;
	}
	return TRUE;
}


void XmlTable::undo() {
	if (undoIndex <= 0) return;
	QFile undoFile(tempFilePath("undo", tmpFileIndex, undoIndex));
	undoIndex = undoIndex - 1;
	QDomDocument xmldoc;
	bool isXml = xmldoc.setContent(&undoFile);
	if (!isXml) return;
	QDomElement root = xmldoc.documentElement();
	int rpos = root.attribute("rA", "0").toInt();
	int cpos = root.attribute("cA", "0").toInt();
	int rposZ = root.attribute("rZ", "0").toInt();
	int cposZ = root.attribute("cZ", "0").toInt();
	addRedoRecord(rpos, cpos, rposZ, cposZ);
	ensureCellVisible(rpos, cpos);
	QDomNodeList trList = root.childNodes();
	for (uint i = 0, r = 0; i < trList.length(); i++) {
		QDomNode trNode = trList.item(i);
		if (!(trNode.isElement() && trNode.nodeName() == "tr")) continue;
		QDomNodeList tdList = trNode.childNodes();
		for (uint j = 0, c = 0; j < tdList.length(); j++) {
			QDomNode tdNode = tdList.item(j);
			if (!(tdNode.isElement() && tdNode.nodeName() == "td")) continue;
			int row = rpos + r;
			int col = cpos + c;
			if (row < numRows() && col < numCols()) {
				QDomElement currElem = tdNode.toElement();
				setText(row, col, currElem.text());
				XmlTableItem* currXmlItem = xmlItem(row, col);
				if (currXmlItem != NULL) {
					currXmlItem->setSpan( 1,1 );
					currXmlItem->setItemType( itemTypeForString( currElem.attribute("class", "") ) );
					if (col == 0) currXmlItem->setDepth( currElem.attribute("depth", "").toInt() );
				}
				updateCell(row, col);
			}
			c++;
		}
		r++;		
	}
	setModified(TRUE);
	emit undoRecordSizeChanged(undoIndex);
}


void XmlTable::redo() {
	if (redoIndex <= 0) return;
	QFile redoFile(tempFilePath("redo", tmpFileIndex, redoIndex));
	redoIndex = redoIndex - 1;
	QDomDocument xmldoc;
	bool isXml = xmldoc.setContent(&redoFile);
	if (!isXml) return;
	QDomElement root = xmldoc.documentElement();
	int rpos = root.attribute("rA", "0").toInt();
	int cpos = root.attribute("cA", "0").toInt();
	int rposZ = root.attribute("rZ", "0").toInt();
	int cposZ = root.attribute("cZ", "0").toInt();
	addUndoRecord(rpos, cpos, rposZ, cposZ, TRUE);
	ensureCellVisible(rpos, cpos);
	QDomNodeList trList = root.childNodes();
	for (uint i = 0, r = 0; i < trList.length(); i++) {
		QDomNode trNode = trList.item(i);
		if (!(trNode.isElement() && trNode.nodeName() == "tr")) continue;
		QDomNodeList tdList = trNode.childNodes();
		for (uint j = 0, c = 0; j < tdList.length(); j++) {
			QDomNode tdNode = tdList.item(j);
			if (!(tdNode.isElement() && tdNode.nodeName() == "td")) continue;
			int row = rpos + r;
			int col = cpos + c;
			if (row < numRows() && col < numCols()) {
				QDomElement currElem = tdNode.toElement();
				setText(row, col, currElem.text());
				XmlTableItem* currXmlItem = xmlItem(row, col);
				if (currXmlItem != NULL) {
					currXmlItem->setSpan( 1,1 );
					currXmlItem->setItemType( itemTypeForString( currElem.attribute("class", "") ) );
					if (col == 0) currXmlItem->setDepth( currElem.attribute("depth", "").toInt() );
				}
				updateCell(row, col);
			}
			c++;
		}
		r++;		
	}
	setModified(TRUE);
	emit redoRecordSizeChanged(redoIndex);
}


void XmlTable::relayMousePress( int row, int col, int button, const QPoint & mousePos ) {
	if (itemType(row, col) != XmlTableItem::ELEMENT	|| button != Qt::LeftButton	|| !inExpandBox(row, col, mousePos)) 
	{
		return;
	}
	XmlTableItem *curritem = xmlItem(row, col);
	bool ex = !curritem->expanded();
	curritem->setExpanded(ex);
	updateCell(row, col);

	int currDepth = depth(row);
	for (int r = row+1; r <= max_rows; r++) {
		if (depth(r) <= currDepth) {
			if (itemType(r, col) == XmlTableItem::ELEMENT) break;
			else if (depth(r+1) <= currDepth
				&& itemType(r+1, col) == XmlTableItem::ELEMENT 
				&& itemType(r, col+1) == XmlTableItem::ATTR_KEY) break;
		}
		curritem = xmlItem(r, col);
		if (curritem != NULL) {
			curritem->setExpanded(ex);
			updateCell(r, col);
		}
		if (ex) showRow(r);
		else hideRow(r);
	}
}

void XmlTable::setNumRows( int r ) {
	Q3Table::setNumRows(r);
	setLeftMargin( fontMetrics().width("E") * QString::number(r).length() );
}


void XmlTable::adjustColumn( int col ) {
	Q3Table::adjustColumn( col );
	if (columnWidth(col) < minColumnWidth) {
		setColumnWidth( col, minColumnWidth );
	}
}


void XmlTable::setText( int r, int c, const QString & text ) {
	XmlTableItem *ti = xmlItem(r, c);
	if (ti != NULL) {
		if (text != ti->text()) setModified(TRUE);
		ti->setText(text);
		if (c == 0) {
			ti->setItemType(XmlTableItem::ELEMENT);
		}
		return;
	}

	int guessedDepth = 0;
	XmlTableItem::NodeType guessedType = XmlTableItem::EMPTY,
		aboveType = itemType(r-1, c), 
		leftType = XmlTableItem::EMPTY;
	for (int i = c-1; i >= 0 && leftType == XmlTableItem::EMPTY; i--) {
		leftType = itemType(r, i);
	}	

	if (text.isEmpty()) {
		return;
	}
	else if (c == 0) {
		guessedType = XmlTableItem::ELEMENT;
		guessedDepth = previousDepth(r);
		if (guessedDepth > 1) guessedDepth--;
	}
	else if (leftType == XmlTableItem::ATTR_KEY || leftType == XmlTableItem::ATTR_VAL) {
		guessedType = leftType;
	}
	else if (leftType == XmlTableItem::ELEMENT) {
		guessedType = XmlTableItem::ATTR_VAL;		
	}
	else if (leftType == XmlTableItem::EMPTY && c == 1) {
		guessedType = XmlTableItem::ATTR_KEY;		
	}
	else if (aboveType == XmlTableItem::ATTR_KEY || leftType == XmlTableItem::ATTR_VAL) {
		guessedType = XmlTableItem::ATTR_VAL;
	}

	XmlTableItem* xitem = new XmlTableItem(guessedType, this, Q3TableItem::OnTyping, text);
	if (c == 0) xitem->setDepth(guessedDepth);
	setItem(r, c, xitem);
	if (!text.isEmpty()) setModified(TRUE);
	if (guessedType == XmlTableItem::ELEMENT) {
		updateColumn(0);
	}
}


void XmlTable::setItem( int r, int c, Q3TableItem *i ) {
	//addUndoRecord(r, c, r, c);
	if (r >= numRows()-1) setNumRows(r+500);
	if (c >= numCols()-1) setNumCols(c+20);
	Q3Table::setItem( r, c, i );
	if (r > max_rows) max_rows = r;
	if (c > max_cols) max_cols = c;
	setModified(TRUE);
}


void XmlTable::expandDoc() {
	setNumRows(numRows()+100);
	setNumCols(numCols()+10);
}


void XmlTable::trimDoc() {
	setNumRows(max_rows+2);
	setNumCols(max_cols+1);
}



void XmlTable::touch(bool val) {
	if (val == TRUE) setModified(TRUE);
}

void XmlTable::setModified( bool val ) {
	isMod = val;
	emit modificationChanged(val);
}


bool XmlTable::isModified() const {
	return isMod;
}




QString XmlTable::getKey(int r, int c) {
	int currRow;
	XmlTableItem* currXmlItem;
	for (currRow = r-1; currRow >= 0; currRow--) {
		currXmlItem = xmlItem(currRow, c);
		if (currXmlItem == NULL) continue;
		switch (currXmlItem->itemType()) {
			case XmlTableItem::ELEMENT:
				currRow = -1;
				break;
			case XmlTableItem::ATTR_KEY: 
				{
					QString retval = currXmlItem->text();
					bool wasChanged;
					retval = fixXmlNameString(retval, &wasChanged);
					if (wasChanged) {
						currXmlItem->setText(retval);
						emit message(
							tr("Attribute name had illegal characters that were replaced with \"_\" in the save."),
							tr("Row: %1, Column: %2").arg(currRow+1).arg(c+1),
							FALSE, TRUE, QMessageBox::Warning
							);
					}
					return retval;
				}
				break;
			default:
				break;
		}
	}
	return QString("C%1").arg(c);
}


void XmlTable::save(QTextStream & out, const QString &docType) {
	out << QString("<?xml version=\"%1\" encoding=\"%2\"?>").arg(xmlVersion).arg(XmlTable::outputEncoding);
	if (!docType.isEmpty()) out << endl << docType;

	bool firstElemEncountered = FALSE;
	for (int r = 0; r <= maxRows(); r++) {
		XmlTableItem* item = xmlItem(r, 0);
		if (item != NULL && item->itemType() == XmlTableItem::ELEMENT) {
			if (!firstElemEncountered) {
				item->setDepth(0);
				firstElemEncountered = TRUE;
			}
			else {
				if (item->depth() <= 1) item->setDepth(1);
			}
		}
	}

	printDocument(out, 0,0, max_rows+1, max_cols+1);
	setModified(FALSE);
}


//returns length of range between current element and next
int XmlTable::getElementRange(int row, int currDepth, int lenRow) {
	int currRow = row+1;
	for (int i = 1; i < lenRow; currRow++, i++) {
		XmlTableItem* item = xmlItem(currRow, 0);
		if (item == NULL) continue;
		if (item->itemType() == XmlTableItem::ELEMENT) {
			if (item->depth() <= currDepth) return currRow - row;
		}
	}
	return currRow - row;
}




QString XmlTable::encodeToHtml(const QString &str, int cellWidth, XmlTableItem::NodeType type) {
	QFontMetrics metric(XmlTable::style(type)->font());
	QString currLine;
	QString xml;
	QChar currChar;
	for (int i = 0; i < str.length(); i++) {
		currChar = str.at(i);
		currLine += currChar;
		if ( cellWidth > 0 
			&& metric.width(currLine, currLine.length()) > cellWidth) {
				currLine = currChar;
				xml += "<br>";
		}
		switch (currChar.toAscii()) {
			case '&': xml += "&amp;"; break;
			case '<': xml += "&lt;"; break;
			case '>': xml += "&gt;"; break;
			case '\'': xml += "&apos;"; break;
			case '\"': xml += "&quot;"; break;
			case '\n': 
				xml += "<br>"; 
				currLine = "";
				break;
			default: xml += currChar; break;
		}
	}
	return xml;
}


bool XmlTable::printDocument(QTextStream & out, int startRow, int startCol, int lenRow, int lenCol, bool* elementClosed) {
	bool childPrinted = FALSE;
	for (int currRow = startRow, i = 0; i < lenRow; currRow++, i++) {
		if (itemType(currRow, startCol) == XmlTableItem::ELEMENT) {
			if(elementClosed != NULL && !(*elementClosed)) {out << ">"; *elementClosed = TRUE; }
			printIndent(out, depth(currRow));
			int range = getElementRange(currRow, depth(currRow), lenRow-i);
			printElement(out, currRow, startCol, range, lenCol);
			childPrinted = childPrinted || TRUE;
			i += range - 1;
			currRow += range - 1;
			continue;
		}
		int currCol = startCol + 1;
		switch ( itemType(currRow, currCol) ) {
			case XmlTableItem::COMMENT:
				if(elementClosed != NULL && !(*elementClosed)) {out << ">"; *elementClosed = TRUE; }
				printIndent(out, depth(currRow)+1);
				out << "<!-- " << encodeComment( text(currRow, currCol) ) << " -->";
				childPrinted = childPrinted || TRUE;
				break;
			case XmlTableItem::TEXT:
				if(elementClosed != NULL && !(*elementClosed)) {out << ">"; *elementClosed = TRUE; }
				printIndent(out, depth(currRow)+1);
				out << encodeToXml(text(currRow, currCol));
				childPrinted = childPrinted || FALSE;
				break;
			case XmlTableItem::CDATA: {
				if(elementClosed != NULL && !(*elementClosed)) {out << ">"; *elementClosed = TRUE; }
				printIndent(out, depth(currRow)+1);
				QString cdataText = text(currRow, currCol);
				int numContained = cdataText.count("]]>");
				if (numContained > 0) {
					emit message(
						tr("CDATA section contains %1 premature case(s) of the termination sequence: ]]>").arg(numContained),
						tr("Row: %1, Column: %2").arg(currRow+1).arg(currCol+1),
						FALSE, TRUE,
						QMessageBox::Critical
						);
				}
				out << "<![CDATA[" << cdataText << "]]>";
				childPrinted = childPrinted || FALSE;
				break;
									  }
			default:
				break;
		}
	}
	return childPrinted;
}


void XmlTable::printIndent(QTextStream &out, int len) {
	out << endl;
	for (int i = 1; i < len; i++) out << indent;
}


QString XmlTable::fixXmlNameString(const QString &val, bool* wasChanged) {
	QString name = val.stripWhiteSpace();
	*wasChanged = FALSE;
	if (name.length() == 0) {
		*wasChanged = TRUE;
		return "_";
	}
	QString retval;
	for (int i = 0; i < name.length(); i++) {
		QChar currChar = name.at(i);
		if (i == 0) {
			if (isLetter(currChar) || currChar == '_' || currChar == ':') {
				retval += currChar;
			}
			else {
				*wasChanged = TRUE;
				retval += '_';
			}
		}
		else {
			if (isLetter(currChar) || isDigit(currChar) 
				|| currChar == '.' || currChar == '-' || currChar == '_' || currChar == ':'
				|| isCombiningChar(currChar) || isExtender(currChar)) {
					retval += currChar;
			}
			else {
				*wasChanged = TRUE;
				retval += '_';
			}
		}
	}
	return retval;
}






void XmlTable::printElement(QTextStream & out, int startRow, int startCol, int lenRow, int lenCol) {
	QString elemName = text(startRow, startCol);
	bool wasChanged;
	elemName = fixXmlNameString(elemName, &wasChanged);
	if (wasChanged) {
		setText(startRow, startCol, elemName);
		emit message(
			tr("Element had illegal characters that were replaced with \"_\" in the save."),
			tr("Row: %1, Column: %2").arg(startRow+1).arg(startCol+1),
			FALSE, TRUE, QMessageBox::Warning
			);
	}
	//TODO:error msg if element name uses illegal characters
	out << "<" << elemName;
	bool elementClosed = FALSE;
	bool childPrinted = FALSE;
	printAttr(out, startRow, startCol+1, lenCol-1, &elementClosed);
	if (lenRow > 1 && lenCol > 1) {
		childPrinted = printDocument(out, startRow+1, startCol, lenRow-1, lenCol, &elementClosed);
		if (childPrinted) printIndent(out, depth(startRow));
	}
	if(elementClosed) {
		out << "</" << elemName << ">";
	}
	else {
		out << "/>";
	}
}


void XmlTable::printAttr(QTextStream & out, int startRow, int startCol, int lenCol, bool* elementClosed) {
	QString currKey;
	//TODO:error msg if attr name uses illegal characters
	for (int currCol = startCol, i = 0; i < lenCol; currCol++, i++) {
		switch ( itemType(startRow, currCol) ) {
			case XmlTableItem::ATTR_VAL:
				currKey = getKey(startRow, currCol);
				out << " " << currKey << "=\"" << encodeToXml(text(startRow, currCol)) << "\"";
				break;
			case XmlTableItem::TEXT:
				if(elementClosed != NULL && !(*elementClosed)) {out << ">"; *elementClosed = TRUE; }
				out << encodeToXml(text(startRow, currCol));
				return;
			case XmlTableItem::CDATA:
				if(elementClosed != NULL && !(*elementClosed)) {out << ">"; *elementClosed = TRUE; }
				out << "<![CDATA[" << text(startRow, currCol) << "]]>";
				return;
			default:
				break;
		}
	}
}


XmlTableItem* XmlTable::xmlItem(int r, int c) {
	Q3TableItem *currItem = item(r, c);
	if (currItem == NULL) return NULL;
	return (XmlTableItem*)currItem;
}




void XmlTable::updateAll() {
	for (int c = 0; c <= max_cols; c++) {
		updateColumn(c);
	}	
}


void XmlTable::updateColumn(int col) {
	for (int r = 0; r <= max_rows; r++) {
		updateCell(r, col);
	}
}


void XmlTable::updateRow(int row) {
	//for (int c = 0; c <= max_cols; c++) {
	for (int c = 0; c < numCols(); c++) {
		updateCell(row, c);
	}	
}


void XmlTable::adjustAll() {
	//adjust column sizes, span text cells to take remainder of row
	XmlTableItem *currXmlItem;
	int c, r;
	for (c = 0; c <= max_cols; c++) {
		for (r = 0; r <= max_rows; r++) {
			currXmlItem = xmlItem(r, c);
			if (currXmlItem == NULL) continue;
			if (c != currXmlItem->col()) continue;
			if (currXmlItem->isSpanCell() ) {
				currXmlItem->setSpan( 1, max_cols+1 - currXmlItem->col() );
			}
			else {
				currXmlItem->setSpan(1, 1);
			}
		}
	}
	for (c = 0; c <= max_cols; c++) {
		for (r = 0; r <= max_rows; r++) {
			currXmlItem = xmlItem(r, c);
			if (currXmlItem == NULL) continue;
			if ( !XmlTable::style(currXmlItem->itemType())->isWordWrap() ) {
				currXmlItem->setSizeHint();
			}
		}
	}


	for (c = 0; c <= max_cols; c++) {
		for (r = 0; r <= max_rows; r++) {
			currXmlItem = xmlItem(r, c);
			if (currXmlItem == NULL) continue;
			if ( XmlTable::style(currXmlItem->itemType())->isWordWrap() ) {
				currXmlItem->setSizeHint();
			}
		}
	}
	for (c = 0; c <= max_cols; c++) adjustColumn(c);
	for (c = 0; c <= max_cols; c++) {
		for (r = 0; r <= max_rows; r++) {
			currXmlItem = xmlItem(r, c);
			if (currXmlItem == NULL) continue;
			if ( XmlTable::style(currXmlItem->itemType())->isWordWrap() ) {
				currXmlItem->setSizeHint();
			}
		}
	}
	for (r = 0; r <= max_rows; r++) adjustRow(r);
}


bool XmlTable::setContent(QIODevice *dev) {
	//clear all existing data
	setUpdatesEnabled(FALSE);
	for (int r = 0; r <= max_rows; r++) {
		for (int c = 0; c <= max_cols; c++) {
			clearCell(r, c);
		}
	}
	resetUndo();
	resetRedo();
	QDomDocument xmldoc;
	QString errorMsg;
	int errorLine, errorColumn;
	bool retval = xmldoc.setContent(dev, &errorMsg, &errorLine, &errorColumn);
	if (!retval) {
		emit message(
			tr("XML read error: %1").arg(errorMsg),
			tr("Line: %1, Column: %2").arg(errorLine).arg(errorColumn),
			FALSE, TRUE, QMessageBox::Critical
			);
		updateAll();
		return retval;
	}

	QDomDocumentType dtd = xmldoc.doctype();
	QString docType;
	QTextStream dtds( &docType, QIODevice::WriteOnly );
	dtd.save(dtds, 0);
	docType = docType.stripWhiteSpace();
	docTypeName = dtd.name();
	emit dtdLoaded(docType);

	QDomElement root = xmldoc.documentElement();
	max_rows = 0; max_cols = 0;
	int lastLevel = 0;

	QStringList *lastAttrList = NULL;
	layoutNode(max_rows, 0, &root, &lastAttrList, &lastLevel);
	if (lastAttrList != NULL) delete lastAttrList;

	if (max_rows > 1) max_rows--;
	setUpdatesEnabled(TRUE);

	adjustAll();
	setModified(FALSE);
	setCurrentCell(0, 0);
	return retval;
}


void XmlTable::insertRows( int r, int n ) {
	if (r <= max_rows) {
		max_rows++;
		//if (max_rows >= numRows()) setNumRows(numRows()+500);
		Q3Table::insertRows(r, n);
	}
	else {
		setNumRows(numRows()+1);
		return;
	}
}


void XmlTable::layoutNode(int &row, int level, QDomNode *node, QStringList **lastAttrList, int *lastLevel) {
	if (node->isElement()) {
		QDomElement elem = node->toElement();
		layoutData(row, level, &elem, lastAttrList, lastLevel);
		if (!node->hasChildNodes()) return;

		QDomNodeList list = node->childNodes();
		QDomNode currNode;
		//skip 1st TEXT or CDATA section
		bool skippedText = FALSE;
		for (uint i = 0; i < list.length(); i++) {
			currNode = list.item(i);
			if (!skippedText) switch (currNode.nodeType()) {
				case QDomNode::TextNode:
				case QDomNode::CDATASectionNode:
					skippedText = TRUE;
					continue;
				default:
					skippedText = TRUE;
			}
			layoutNode(row, level+1, &currNode, lastAttrList, lastLevel);
		}
	}
	else {
		layoutText(row, 1, node);
	}
}




bool XmlTable::isNewNodeType(QStringList *newAttrs, QStringList *oldAttrs) {
	if (oldAttrs == NULL || newAttrs == NULL) return TRUE;
	if (newAttrs->empty()) return FALSE;
	return (*newAttrs) != (*oldAttrs);
}


void XmlTable::layoutData(int &row, int level, QDomElement *elem, QStringList **lastAttrList, int *lastLevel) {
	int curr_col = 0;
	uint i;
	QDomNamedNodeMap map = elem->attributes();

	QStringList *attrKeyList = new QStringList();
	for (i = 0; i < map.length(); i++) {
		attrKeyList->append(map.item(i).nodeName());
	}
	if (XmlTable::sortAttr) attrKeyList->sort();

	if (isNewNodeType(attrKeyList, *lastAttrList)) {
		clearCell(row, curr_col++);
		for (QStringList::iterator it = attrKeyList->begin(); it != attrKeyList->end(); ++it) {
			setItem(row, curr_col++,
				new XmlTableItem(XmlTableItem::ATTR_KEY, this, Q3TableItem::OnTyping, *it)
				);
		}
		row++;
		curr_col = 0;
	}
	XmlTableItem* elemItem = new XmlTableItem(XmlTableItem::ELEMENT, this, Q3TableItem::OnTyping, elem->nodeName());
	if (curr_col == 0) elemItem->setDepth(level);
	setItem(row, curr_col++, elemItem);

	for (QStringList::iterator it = attrKeyList->begin(); it != attrKeyList->end(); ++it) {
		setItem(row, curr_col++,
			new XmlTableItem(XmlTableItem::ATTR_VAL, this, Q3TableItem::OnTyping, elem->attribute(*it, ""))
			);
	}

	QDomNodeList list = elem->childNodes();
	QDomNode currNode;
	//print 1st TEXT or CDATA section only
	bool found = FALSE, wasAdjusted = FALSE;
	for (i = 0; i < list.length() && !found; i++) {
		currNode = list.item(i);
		switch (currNode.nodeType()) {
			case QDomNode::TextNode:
			case QDomNode::CDATASectionNode:
				found = TRUE;
				layoutText(row, curr_col++, &currNode);
				wasAdjusted = TRUE;
				continue;
			default:
				found = TRUE;
		}
	}

	if (!wasAdjusted) adjustRow(row++);

	if (lastAttrList != NULL && *lastAttrList != NULL) delete (*lastAttrList);
	*lastAttrList = attrKeyList;

	*lastLevel = level;
	if (curr_col > max_cols) max_cols = curr_col;
}


void XmlTable::layoutText(int &row, int level, QDomNode *node) {
	XmlTableItem::NodeType ntype;
	switch (node->nodeType()) {
		case QDomNode::TextNode:
			ntype = XmlTableItem::TEXT;
			break;
		case QDomNode::CDATASectionNode:
			ntype = XmlTableItem::CDATA;
			break;
		case QDomNode::CommentNode:
			ntype = XmlTableItem::COMMENT;
			break;
		default:
			ntype = XmlTableItem::TEXT;
			break;
	}

	QString val = node->nodeValue().stripWhiteSpace();
	XmlTableItem *item = new XmlTableItem(ntype, this, Q3TableItem::OnTyping, val);
	setItem(row, level, item);
	adjustRow(row++);

	if (level > max_cols) max_cols = level;
}


QString XmlTable::html(int topRow, int leftCol, int bottomRow, int rightCol) {
	QString str = "<html><body>";
	XmlTableItem* currItem;
	str += "<table border='1' cellpadding='0' cellspacing='0'>";
	for (int r = topRow; r <= bottomRow; r++) {
		str += "\n<tr>";
		for (int c = leftCol; c <= rightCol; c++) {
			currItem = xmlItem(r, c);
			if (currItem == NULL) {
				str += "<td/>";
				continue;
			}
			XmlTableItem::NodeType currType = currItem->itemType();
			bool isPre = (currType == XmlTableItem::TEXT || currType == XmlTableItem::CDATA);
			XmlTableStyle* currStyle = style(currType);
			QFont currFont = currStyle->font();
			str += QString("<td colspan='%1' bgcolor='%2' align='%3' valign='%4'>")
				.arg( currItem->colSpan() )
				.arg( currStyle->bgColor().name() )
				.arg( currStyle->alignText() )
				.arg( currStyle->valignText() );
			if (isPre) str += "<pre>"; 
			str += QString("<font color='%1' face='%2' size='%3'>")
				.arg( currStyle->fgColor().name() )
				.arg( currFont.family() )
				.arg( /*currFont.pointSize()*/0 );
			if (currFont.bold()) str += "<b>";
			if (currFont.italic()) str += "<i>";
			str += encodeToHtml(text(r, c));
			if (currFont.italic()) str += "</i>";
			if (currFont.bold()) str += "</b>";
			if (currItem != NULL && currItem->isSpanCell()) c = rightCol+1;
			str += "</font>";
			if (isPre) str += "</pre>"; 
			str += "</td>";
		}
		str += "</tr>";
	}
	str += "\n</table></body></html>";
	return str;
}


QString XmlTable::html2(int topRow, int leftCol, int bottomRow, int rightCol) {
	int maxDepth = 1;
	for (int r = topRow; r <= bottomRow; r++) {
		maxDepth = (maxDepth > depth(r)) ? maxDepth : depth(r);
	}
	QString str = "<html><body>";
	XmlTableItem* currItem;
	str += "<table border='1' frame='void' cellpadding='0' cellspacing='0'>";
	for (int r = topRow; r <= bottomRow; r++) {
		str += "\n<tr>";
		for (int c = leftCol; c <= rightCol; c++) {
			currItem = xmlItem(r, c);
			if (currItem == NULL) {
				if (c == 0) {
					for (int i = 1; i <= depth(r); i++) {
						str += QString("<td width='%1'>&nbsp;</td>").arg(indentWidth);
					}
					str += QString("<td colspan='%1'>&nbsp;</td>").arg(maxDepth - depth(r) );
				}
				else str += "<td>&nbsp;</td>";
				continue;
			}
			XmlTableItem::NodeType currType = currItem->itemType();
			bool isPre = (currType == XmlTableItem::TEXT || currType == XmlTableItem::CDATA);
			XmlTableStyle* currStyle = style(currType);
			QFont currFont = currStyle->font();
			if (currType == XmlTableItem::ELEMENT) {
				for (int i = 1; i < currItem->depth(); i++) {
					str += QString("<td width='%1'>&nbsp;</td>").arg(indentWidth);
				}
				str += QString("<td colspan='%1' bgcolor='%2' align='%3' valign='%4'>")
					.arg( maxDepth - currItem->depth() +1 )
					.arg( currStyle->bgColor().name() )
					.arg( currStyle->alignText() )
					.arg( currStyle->valignText() );
			}
			else if (!currItem->isSpanCell()) {
				str += QString("<td bgcolor='%1' align='%2' valign='%3'>")
					.arg( currStyle->bgColor().name() )
					.arg( currStyle->alignText() )
					.arg( currStyle->valignText() );
			}
			else {
				str += QString("<td colspan='%1' bgcolor='%2' align='%3' valign='%4'>")
					.arg( currItem->colSpan() )
					.arg( currStyle->bgColor().name() )
					.arg( currStyle->alignText() )
					.arg( currStyle->valignText() );
			}

			if (isPre) str += "<pre>"; 

			str += QString("<font color='%1' face='%2' size='%3'>")
				.arg( currStyle->fgColor().name() )
				.arg( currFont.family() )
				.arg( /*currFont.pointSize()*/0 );
			if (currFont.bold()) str += "<b>";
			if (currFont.italic()) str += "<i>";

			str += encodeToHtml(text(r, c), currItem->spanWidth(), currType);

			if (currFont.italic()) str += "</i>";
			if (currFont.bold()) str += "</b>";
			if (currItem != NULL && currItem->isSpanCell()) c = rightCol+1;
			str += "</font>";

			if (isPre) str += "</pre>"; 
			str += "</td>";
		}
		str += "</tr>";
	}
	str += "\n</table></body></html>\n";
	return str;
}


QString XmlTable::html_css(int topRow, int leftCol, int bottomRow, int rightCol) {
	int maxDepth = 1;
	for (int r = topRow; r <= bottomRow; r++) {
		//maxDepth = maxDepth >? depth(r);
		maxDepth = (maxDepth > depth(r)) ? maxDepth : depth(r);
	}

	QString str = "<html><head><style type='text/css'>\n"
		"body { margin: 0 0 0 0; } \n";

	for (int type = XmlTableItem::EMPTY; type <= XmlTableItem::COMMENT; type++) {
		XmlTableStyle* currStyle = styleVector[type];
		QString name = currStyle->name();
		QFont font = currStyle->font();
		QString style = font.italic() ? "italic":"normal";
		str += QString(".tx_%1 { \n"
			"color:%2; background:%3; \n"
			"text-align:%4; vertical-align:%5; \n"
			"font-family:%6; font-size:%7pt; font-weight:%8; font-style:%9; \n")
			.arg(currStyle->name())
			.arg(currStyle->fgColor().name())
			.arg(currStyle->bgColor().name())
			.arg(currStyle->alignText())
			.arg(currStyle->valignText())
			.arg(font.family())
			.arg(font.pointSize())
			.arg(font.weight()*8)
			.arg(style);
		switch (type) {
			case XmlTableItem::EMPTY:
				str += "border-top:thin dashed gray; border-right:thin dashed gray; \n";
				break;
			case XmlTableItem::ELEMENT:
				str += "border-top:thin solid black; border-right:thin solid black; border-left:thin solid black; \n";
				break;
			default:
				str += "border-top:thin solid black; border-right:thin solid black; \n";
		}
		str += "}\n";
	}
	str += QString(".tx_emptyA { \n"
		"border-top:thin dashed gray; border-left:thin dashed gray; \n"
		"}\n"
		);
	str += QString(".tx_emptyZ { "
		"border-top:thin dashed gray; border-left:thin dashed gray; border-right:thin solid black; "
		"}\n"
		);

	str += "</style></head><body>";
	XmlTableItem* currItem;
	str += "<table cellpadding='0' cellspacing='0'>";
	for (int r = topRow; r <= bottomRow; r++) {
		str += "\n<tr>";
		for (int c = leftCol; c <= rightCol; c++) {
			currItem = xmlItem(r, c);
			if (currItem == NULL) {
				if (c == 0) {
					for (int i = 1; i <= depth(r); i++) {
						str += QString("<td class='tx_emptyA' width='%1'>&nbsp;</td>").arg(indentWidth);
					}
					str += QString("<td class='tx_emptyZ' colspan='%1'>&nbsp;</td>").arg(maxDepth - depth(r) );
				}
				else str += "<td class='tx_empty'>&nbsp;</td>";
				continue;
			}
			XmlTableItem::NodeType currType = currItem->itemType();
			bool isPre = (currType == XmlTableItem::TEXT || currType == XmlTableItem::CDATA);
			XmlTableStyle* currStyle = style(currType);
			QFont currFont = currStyle->font();
			if (currType == XmlTableItem::ELEMENT) {
				for (int i = 1; i < currItem->depth(); i++) {
					str += QString("<td class='tx_emptyA' width='%1'>&nbsp;</td>").arg(indentWidth);
				}
				str += QString("<td class='tx_%1' colspan='%2'>")
					.arg( currStyle->name() )
					.arg( maxDepth - currItem->depth() +1);
			}
			else if (!currItem->isSpanCell()) {
				str += QString("<td class='tx_%1'>").arg( currStyle->name() );
			}
			else {
				str += QString("<td class='tx_%1' colspan='%2'>")
					.arg( currStyle->name() )
					.arg( currItem->colSpan() );
			}

			if (isPre) str += "<pre>"; 
			str += encodeToHtml(text(r, c), currItem->spanWidth(), currType);
			if (currItem != NULL && currItem->isSpanCell()) c = rightCol+1;
			if (isPre) str += "</pre>"; 
			str += "</td>";
		}
		str += "</tr>";
	}
	str += "\n</table></body></html>\n";
	return str;
}


/*
<table r='x' c='y'>
<tr><td class='type' depth='d'>text</td>
</tr>
</undo>

*/


void XmlTable::copy(bool isTextOnly) {
	if (isTextOnly) {
		QString str;
		int numSel = numSelections();
		if (numSel <= 0) {
			selectCells(currentRow(), currentColumn(), currentRow(), currentColumn());
			numSel = numSelections();
		}
		XmlTableItem* currItem;
		for (int k = 0; k < numSel; k++) {
			Q3TableSelection sel = selection(k);
			if (sel.isEmpty()) continue;
			if (k > 0) str += "\n";
			for (int r=sel.topRow(); r <= sel.bottomRow(); r++) {
				if (r > sel.topRow()) str += "\n";
				for (int c = sel.leftCol(); c <= sel.rightCol(); c++) {
					if (c > sel.leftCol()) str += "\t";
					str += text(r, c);
					currItem = xmlItem(r, c);
					if (currItem != NULL && currItem->isSpanCell()) c = sel.rightCol()+1;
				}
			}
		}
		QApplication::clipboard()->setText(str);
	}//IF
	else {
		QString str = "<table>";
		int numSel = numSelections();
		if (numSel <= 0) {
			selectCells(currentRow(), currentColumn(), currentRow(), currentColumn());
			numSel = numSelections();
		}
		XmlTableItem* currItem;
		for (int k = 0; k < numSel; k++) {
			Q3TableSelection sel = selection(k);
			if (sel.isEmpty()) continue;
			for (int r=sel.topRow(); r <= sel.bottomRow(); r++) {
				str += "\n<tr>";
				for (int c = sel.leftCol(); c <= sel.rightCol(); c++) {
					currItem = xmlItem(r, c);
					if (currItem != NULL) {
						str += QString("<td class='%1' depth='%2'>")
							.arg( style(currItem->itemType())->name() )
							.arg( currItem->depth()-1 );
					}
					else str += "<td>";
					str += encodeToXml(text(r, c));
					if (currItem != NULL && currItem->isSpanCell()) c = sel.rightCol()+1;
					str += "</td>";
				}
				str += "</tr>";
			}
		}
		str += "\n</table>";
		QApplication::clipboard()->setText(str);
	}//ELSE
}


void XmlTable::copyXml() {
	QString str;
	int numSel = numSelections();
	if (numSel <= 0) {
		selectCells(currentRow(), currentColumn(), currentRow(), currentColumn());
		numSel = numSelections();
	}
	QTextStream ts( &str, QIODevice::WriteOnly );
	for (int k = 0; k < numSel; k++) {
		Q3TableSelection sel = selection(k);
		if (sel.isEmpty()) continue;
		if (k > 0) ts << endl;
		printDocument(ts, sel.topRow(), 0, sel.bottomRow()-sel.topRow()+1, max_cols+1);
	}
	QApplication::clipboard()->setText(str);
}


void XmlTable::shiftLeft() {
	int numSel = numSelections();
	if (numSel <= 0) selectCells(currentRow(), currentColumn(), currentRow(), currentColumn());
	Q3TableSelection sel = selection(0);
	if (sel.isEmpty()) return;
	//if (sel.leftCol() == 0) return;
	setModified(TRUE);
	if (sel.leftCol() == 0) {
		addUndoRecord(sel.topRow(), 0, sel.bottomRow(), 0);
		for (int r=sel.topRow(); r <= sel.bottomRow(); r++) {
			XmlTableItem *currItem = xmlItem(r, 0);
			if (currItem != NULL) currItem->decreaseDepth();
			//updateCell(r, 0);
		}
		updateColumn(0);
		return;
	}
	else if (sel.leftCol() == 1) {
		return;
	}
	int top = sel.topRow(), bottom = sel.bottomRow(), left = sel.leftCol(), right = sel.rightCol();
	addUndoRecord(top, 1, bottom, numCols()-1);
	int r, c;
	for (r=sel.topRow(); r <= sel.bottomRow(); r++) {
		for (c = sel.leftCol(); c < numCols(); c++) {
			Q3TableItem* currItem = item(r, c);
			if (currItem != NULL) currItem->setSpan(1, 1);
		}
		if (sel.leftCol() > 0) clearCell(r, sel.leftCol()-1);
		for (c = sel.leftCol(); c < numCols(); c++) {
			if (c > 0) swapCells(r, c, r, c-1);
		}
		updateRow(r);
	}

	if (left > 0) {left--; right--;}
	clearSelection(TRUE);
	setCurrentCell(top, left);
	selectCells( top, left, bottom, right );
	//	adjustAll();
}


void XmlTable::shiftRight() {
	int numSel = numSelections();
	if (numSel <= 0) selectCells(currentRow(), currentColumn(), currentRow(), currentColumn());
	Q3TableSelection sel = selection(0);
	if (sel.isEmpty()) return;
	setModified(TRUE);
	if (sel.leftCol() == 0) {
		addUndoRecord(sel.topRow(), 0, sel.bottomRow(), 0);
		for (int r=sel.topRow(); r <= sel.bottomRow(); r++) {
			XmlTableItem *currItem = xmlItem(r, 0);
			if (currItem != NULL) currItem->increaseDepth();
			//updateCell(r, 0);
		}
		updateColumn(0);
		return;
	}
	else if (sel.rightCol() == numCols()-1) {
		return;
	}
	int top = sel.topRow(), bottom = sel.bottomRow(), left = sel.leftCol(), right = sel.rightCol();
	addUndoRecord(top, 1, bottom, numCols()-1);
	int r, c;
	for (r=sel.topRow(); r <= sel.bottomRow(); r++) {
		for (c = sel.leftCol(); c < numCols(); c++) {
			Q3TableItem* currItem = item(r, c);
			if (currItem != NULL) currItem->setSpan(1, 1);
		}
		if ( !text(r, numCols()-1).isEmpty() ) {
			setNumCols( numCols()+1 );
		}
		for (c = numCols()-1; c > sel.leftCol(); c--) {
			swapCells(r, c, r, c-1);
		}
		updateRow(r);

	}

	if (right < numCols()-1) {left++; right++;}
	clearSelection(TRUE);
	setCurrentCell(top, left);
	selectCells( top, left, bottom, right );
	//adjustAll();
}




void XmlTable::paste() {
	int rpos, cpos;
	if (numSelections() == 0) {
		selectCells(currentRow(), currentColumn(), currentRow(), currentColumn());
	}
	Q3TableSelection sel = selection(0);
	rpos = sel.topRow(); cpos = sel.leftCol();

	QString str = QApplication::clipboard()->text();
	QDomDocument xmldoc;
	bool isXml = xmldoc.setContent(str);

	int minX = numCols(), minY = numRows(), maxX = 0, maxY = 0;
	if (isXml) {
		QDomElement root = xmldoc.documentElement();
		QDomNodeList trList = root.childNodes();
		for (uint i = 0, r = 0; i < trList.length(); i++) {
			QDomNode trNode = trList.item(i);
			if (!(trNode.isElement() && trNode.nodeName() == "tr")) continue;
			QDomNodeList tdList = trNode.childNodes();
			for (uint j = 0, c = 0; j < tdList.length(); j++) {
				QDomNode tdNode = tdList.item(j);
				if (!(tdNode.isElement() && tdNode.nodeName() == "td")) continue;
				int row = rpos + r;
				int col = cpos + c;
				if (row < numRows() && col < numCols()) {
					//minX = minX <? col;
					minX = (minX < col) ? minX : col;
					//minY = minY <? row;
					minY = (minY < row) ? minY : row;
					//maxX = maxX >? col;
					maxX = (maxX > col) ? maxX : col;
					//maxY = maxY >? row;
					maxY = (maxY > row) ? maxY : row;
				}
				c++;
			}
			r++;
		}
		addUndoRecord(minY, minX, maxY, maxX);

		for (uint i = 0, r = 0; i < trList.length(); i++) {
			QDomNode trNode = trList.item(i);
			if (!(trNode.isElement() && trNode.nodeName() == "tr")) continue;
			QDomNodeList tdList = trNode.childNodes();
			for (uint j = 0, c = 0; j < tdList.length(); j++) {
				QDomNode tdNode = tdList.item(j);
				if (!(tdNode.isElement() && tdNode.nodeName() == "td")) continue;
				int row = rpos + r;
				int col = cpos + c;
				if (row < numRows() && col < numCols()) {
					QDomElement currElem = tdNode.toElement();
					setText(row, col, currElem.text());
					XmlTableItem* currXmlItem = xmlItem(row, col);
					if (currXmlItem != NULL) {
						currXmlItem->setSpan( 1,1 );
						currXmlItem->setItemType( itemTypeForString( currElem.attribute("class", "") ) );
						if (col == 0) currXmlItem->setDepth( currElem.attribute("depth", "").toInt() );
					}
					updateCell(row, col);
				}
				c++;
			}
			r++;		
		}
	}//IF
	else {
		QStringList rows = QStringList::split("\n", str, true);
		for (int i = 0, n = rows.size(); i < n; i++) {
			QStringList cols = QStringList::split("\t", rows[i], true);
			for (int j = 0, m = cols.size(); j < m; j++) {
				int row = rpos + i;
				int col = cpos + j;
				if (row < numRows() && col < numCols()) {
					//minX = minX <? col;
					minX = (minX < col) ? minX : col;
					//minY = minY <? row;
					minY = (minY < row) ? minY : row;
					//maxX = maxX >? col;
					maxX = (maxX > col) ? maxX : col;
					//maxY = maxY >? row;
					maxY = (maxY > row) ? maxY : row;
				}
			}
		}
		addUndoRecord(minY, minX, maxY, maxX);

		for (int i = 0, n = rows.size(); i < n; i++) {
			QStringList cols = QStringList::split("\t", rows[i], true);
			for (int j = 0, m = cols.size(); j < m; j++) {
				int row = rpos + i;
				int col = cpos + j;
				if (row < numRows() && col < numCols()) {
					setText(row, col, cols[j]);
					XmlTableItem* currXmlItem = xmlItem(row, col);
					if (currXmlItem != NULL) {
						currXmlItem->setSpan( 1,1 );
						currXmlItem->setItemType(XmlTableItem::EMPTY);
					}
					updateCell(row, col);
				}
			}
		}
	}//ELSE
	setModified(TRUE);
}


void XmlTable::pasteXml() {
	int rpos, cpos;
	if (numSelections() == 0) {
		selectCells(currentRow(), currentColumn(), currentRow(), currentColumn());
	}
	Q3TableSelection sel = selection(0);
	rpos = sel.topRow(); cpos = sel.leftCol();

	QString str = "<paste>";
	str.append( QApplication::clipboard()->text() );
	str.append("</paste>");
	QDomDocument xmldoc;
	bool isXml = xmldoc.setContent(str);
	if (isXml) {
		int currDepth = (rpos > 0) ? depth(rpos-1) : depth(rpos);
		int lastLevel = currDepth;
		QStringList *lastAttrList = NULL;

		QDomElement root = xmldoc.documentElement();
		QDomNodeList trList = root.childNodes();
		for (uint i = 0; i < trList.length(); i++) {
			QDomNode trNode = trList.item(i);
			layoutNode(rpos, currDepth, &trNode, &lastAttrList, &lastLevel);
		}
		if (lastAttrList != NULL) delete lastAttrList;
		resetUndo(); resetRedo();
	}//IF
	else {
		addUndoRecord(rpos, cpos, rpos, cpos);
		setText(rpos, cpos, QApplication::clipboard()->text());
	}
	setModified(TRUE);
}



void XmlTable::del() {
	int numSel = numSelections();
	if (numSel <= 0) {
		selectCells(currentRow(), currentColumn(), currentRow(), currentColumn());
		numSel = numSelections();
	}

	int minX = numCols(), minY = numRows(), maxX = 0, maxY = 0;
	for (int k = 0; k < numSel; k++) {
		Q3TableSelection sel = selection(k);
		if (sel.isEmpty()) continue;
		for (int r=sel.topRow(); r <= sel.bottomRow(); r++) {
			for (int c = sel.leftCol(); c <= sel.rightCol(); c++) {
				//minX = minX <? c;
				minX = (minX < c) ? minX : c;
				//minY = minY <? r;
				minY = (minY < r) ? minY : r;
				//maxX = maxX >? c;
				maxX = (maxX > c) ? maxX : c;
				//maxY = maxY >? r;
				maxY = (maxY > r) ? maxY : r;
			}
		}
	}
	addUndoRecord(minY, minX, maxY, maxX);
	for (int k = 0; k < numSel; k++) {
		Q3TableSelection sel = selection(k);
		if (sel.isEmpty()) continue;
		for (int r=sel.topRow(); r <= sel.bottomRow(); r++) {
			for (int c = sel.leftCol(); c <= sel.rightCol(); c++) {
				clearCell(r, c);
			}
		}
	}
	setModified(TRUE);
}






int XmlTable::depth(int row) {
	XmlTableItem* curr = xmlItem(row, 0);
	if (curr == NULL || curr->itemType() == XmlTableItem::EMPTY) {
		int nextLoc = -1, prevLoc = -1;
		int nextD = nextDepth(row, &nextLoc);
		int prevD = previousDepth(row, &prevLoc);

		int retval;
		//if (nextD != prevD) retval = nextD <? prevD;
		if (nextD != prevD) {retval = ((nextD < prevD) ? nextD : prevD);}
		else retval = nextD - 1;
		return retval;
	}
	return curr->depth();
}


void XmlTable::paintFocus( QPainter * p, const QRect & cr ) {
	QPen pen(focusBorderColor, 3, Qt::DashLine, Qt::FlatCap, Qt::RoundJoin);
	p->setPen( pen );
	p->drawRect( -1,-1, cr.width()+1,cr.height()+1 );
}



bool XmlTable::inExpandBox(int row, int col, const QPoint &p) {
	QFontMetrics metric(XmlTable::style(XmlTableItem::ELEMENT)->font());
	int dy = metric.lineSpacing()/2;
	int y = rowPos(row) + dy;
	int x = columnPos(col) + indentWidth * (depth(row)-1) - indentWidth/2;
	return p.x() >= x-dy && p.x() <= x+dy && p.y() >= y-dy && p.y() <= y+dy;
}


void XmlTable::paintCell(QPainter * p, int row, int col, const QRect & cr, bool selected, const QColorGroup & /*cg*/) {
	p->save();
	XmlTableItem::NodeType type = itemType(row, col);
	QColor bgColor, fgColor;
	if ( selected && !(row == currentRow() && col == currentColumn()) ) {
		fgColor = XmlTable::style(type)->bgColor(), 
			bgColor = XmlTable::style(type)->fgColor();		
	}
	else {
		bgColor = XmlTable::style(type)->bgColor(), 
			fgColor = XmlTable::style(type)->fgColor();
	}
	p->setFont(XmlTable::style(type)->font());
	//p->setTabStops(4); //tr--QT3TO4

	if (type == XmlTableItem::ELEMENT || type == XmlTableItem::EMPTY && col == 0) {
		p->setPen( QPen(borderColor, 2) );
		p->setBrush( Qt::NoBrush );
		p->setBackgroundColor(XmlTable::style(XmlTableItem::EMPTY)->bgColor());

		QFontMetrics metric(XmlTable::style(XmlTableItem::ELEMENT)->font());
		int dh = metric.lineSpacing();
		int dw = indentWidth;
		int w = cr.width(), h = cr.height();

		p->eraseRect(0,0, w, h);
		p->setPen( QPen(borderColor, 1) );

		if (type == XmlTableItem::EMPTY) {
			for (int i = 0; i < depth(row); i++) {
				QPen pen(emptyBorderColor);
				pen.setDashPattern(this->dashPattern);
				pen.setWidth(1);
				p->setPen(pen);
				p->drawLine(1,h-1, dw, h-1);//bottom
				p->drawLine(dw,1, dw, h);//right

				if (displayBranches) {
					p->setPen( QPen(borderColor, 1) );
					int nextD = nextDepth(row);
					int prevD = previousDepth(row);
					if (nextD < prevD && prevD == i+2);
					else if (nextD > prevD && nextD == i+2) p->drawLine(dw/2,0, dw/2,h);
					else if (hasSibling(previousRowWithDepth(row, i+2)) == -1);
					else p->drawLine(dw/2,0, dw/2,h);
				}
				p->translate(dw, 0);
			}
			p->fillRect(0, 0, w - dw*depth(row), h, QBrush(bgColor) );
		}	
		else if (type == XmlTableItem::ELEMENT) {
			for (int i = 1; i < depth(row) && col == 0; i++) {
				QPen pen(emptyBorderColor);
				pen.setWidth(1);
				pen.setDashPattern(this->dashPattern);
				p->setPen(pen);
				p->drawLine(1,h-1, dw, h-1);//bottom
				p->drawLine(dw,1, dw, h);//right

				if (displayBranches) {
					p->setPen( QPen(borderColor, 1) );
					if (i+1 == depth(row)) {
						int nextD = nextDepth(row);
						int thisD = depth(row);
						if (hasSibling(row) < 0) p->drawLine(dw/2,0, dw/2,dh/2);//top only
						else if (i > nextDepth(row)) p->drawLine(dw/2,0, dw/2,dh/2);//top only
						else if (nextD < thisD && thisD == i+1) p->drawLine(dw/2,0, dw/2,dh/2);//top only
						else p->drawLine(dw/2,0, dw/2,h);//top to bottom
						p->drawLine(dw/2,dh/2, dw,dh/2);//right

						if (nextDepth(row) <= depth(row)) {
							if (xmlItem(row, col)->expanded()) {
								p->fillRect(dw/2-1, dh/2-1, 3, 3, QBrush(borderColor) );
							}
							else {
								p->fillRect(dw/2-4, dh/2-4, 9, 9, QBrush(QColor("pink")) );
								p->drawRect(dw/2-4, dh/2-4, 9, 9);
								p->drawLine(dw/2-4, dh/2, dw/2+4, dh/2);
								p->drawLine(dw/2, dh/2-4, dw/2, dh/2+4);
							}
						}
						else {
							if (xmlItem(row, col)->expanded()) {
								p->fillRect(dw/2-4, dh/2-4, 9, 9, QBrush(QColor("lightgreen")) );
								p->drawRect(dw/2-4, dh/2-4, 9, 9);
								p->drawLine(dw/2-4, dh/2, dw/2+4, dh/2);
							}
							else {
								p->fillRect(dw/2-4, dh/2-4, 9, 9, QBrush(QColor("pink")) );
								p->drawRect(dw/2-4, dh/2-4, 9, 9);
								p->drawLine(dw/2-4, dh/2, dw/2+4, dh/2);
								p->drawLine(dw/2, dh/2-4, dw/2, dh/2+4);
							}
						}
					}
					else {
						int nextD = nextDepth(row);
						int prevD = previousDepth(row);
						if (nextD < prevD && prevD == i+1);
						else if (nextD > prevD && nextD == i+1) p->drawLine(dw/2,0, dw/2,h);
						else if (hasSibling(previousRowWithDepth(row, i+1)) == -1);
						else {
							if (i > nextDepth(row)) {
								p->drawLine(dw/2,0, dw/2,dh/2);//top
								p->fillRect(dw/2-1, dh/2-1, 3, 3, QBrush(borderColor) );
							}
							else p->drawLine(dw/2,0, dw/2,h);//to to bottom
						}
					}
				}

				p->translate(dw, 0);
			}
			if (col == 0) w = w - dw*(depth(row)-1);

			p->fillRect(0, 0, w, h, QBrush(bgColor) );
			p->setPen( QPen(fgColor, 1) );
			p->drawText(2, 1, w-5, h-2, XmlTable::style(type)->textFlags(), text(row, col), -1, 0);

			p->setPen( QPen(borderColor, 1) );
			p->drawRect(-1,-1, w, h);
		}
	}
	else {
		XmlTableItem* xitem = xmlItem(row, col);
		int w = cr.width(), h = cr.height();
		if (xitem != NULL) w = xitem->spanWidth();
		p->fillRect(0, 0, w, h, QBrush(bgColor) );
		p->setPen( QPen(fgColor, 1) );
		p->drawText(2, 1, w-4, h-2, XmlTable::style(type)->textFlags(), text(row, col), -1, 0);
		if (type != XmlTableItem::EMPTY) {
			p->setPen( QPen(borderColor, 1) );
			p->drawRect(-1,-1, w, h);
		}
	}
	p->restore();
	if (type == XmlTableItem::EMPTY) {
		int w = cr.width()-1, h = cr.height()-1;
		QPen pen(emptyBorderColor);
		pen.setWidth(1);
		pen.setDashPattern(this->dashPattern);
		p->setPen(pen);
		if (itemType(row+1, col) == XmlTableItem::EMPTY) p->drawLine(1,h, w, h);//bottom
		if (itemType(row, col+1) == XmlTableItem::EMPTY) p->drawLine(w,1, w, h);//right
	}
}



int XmlTable::maxRows() {
	return max_rows;
}
int XmlTable::maxCols() {
	return max_cols;
}





bool XmlTable::isParent(int row) {
	int currDepth = depth(row);
	for (int r = row+1; r <= max_rows; r++) {
		XmlTableItem* itemBelow = xmlItem(r, 0);
		if (itemBelow != NULL && itemBelow->itemType() == XmlTableItem::ELEMENT) {
			if (itemBelow->depth() > currDepth) return TRUE;
			if (itemBelow->depth() <= currDepth) return FALSE;
		}
	}
	return FALSE;
}


bool XmlTable::isLastSibling(int row) {
	int currDepth = depth(row);
	for (int r = row+1; r <= max_rows; r++) {
		XmlTableItem* itemBelow = xmlItem(r, 0);
		if (itemBelow != NULL && itemBelow->itemType() == XmlTableItem::ELEMENT) {
			if (itemBelow->depth() >= currDepth) return FALSE;
			if (itemBelow->depth() < currDepth) return TRUE;
		}
	}
	return TRUE;
}


int XmlTable::hasSibling(int row, int selDepth) {
	int currDepth = selDepth;
	if (currDepth < 0) currDepth = depth(row);
	for (int r = row+1; r <= max_rows; r++) {
		XmlTableItem* itemBelow = xmlItem(r, 0);
		if (itemBelow != NULL && itemBelow->itemType() == XmlTableItem::ELEMENT) {
			if (itemBelow->depth() == currDepth) return r;
			if (itemBelow->depth() < currDepth) return -1;
		}
	}
	return -1;
}


bool XmlTable::isFirstSibling(int row) {
	int currDepth = depth(row);
	for (int r = row-1; r >= 0; r--) {
		XmlTableItem* itemAbove = xmlItem(r, 0);
		if (itemAbove != NULL && itemAbove->itemType() == XmlTableItem::ELEMENT) {
			if (itemAbove->depth() > currDepth) return FALSE;
			if (itemAbove->depth() <= currDepth) return TRUE;
		}
	}
	return TRUE;
}


int XmlTable::nextDepth(int row, int* loc) {
	for (int r = row+1; r <= max_rows; r++) {
		XmlTableItem* itemBelow = xmlItem(r, 0);
		if (itemBelow != NULL && itemBelow->itemType() == XmlTableItem::ELEMENT) {
			if (loc != NULL) *loc = r;
			return itemBelow->depth();
		}
	}
	if (loc != NULL) *loc = -1;
	return 0;
}


int XmlTable::previousDepth(int row, int* loc) {
	for (int r = row-1; r >= 0; r--) {
		XmlTableItem* itemAbove = xmlItem(r, 0);
		if (itemAbove != NULL && itemAbove->itemType() == XmlTableItem::ELEMENT) {
			if (loc != NULL) *loc = r;
			return itemAbove->depth();
		}
	}
	if (loc != NULL) *loc = 0;
	return 0;
}


int XmlTable::previousRowWithDepth(int startRow, int depth) {
	for (int r = startRow-1; r >= 0; r--) {
		XmlTableItem* itemAbove = xmlItem(r, 0);
		if (itemAbove != NULL && itemAbove->itemType() == XmlTableItem::ELEMENT) {
			if (itemAbove->depth() == depth) return r;
		}
	}
	return 0;
}





void XmlTable::setEditingLocation(int r, int c) {
	editingRow = r;
	editingColumn = c;
}


void XmlTable::setCellContentFromEditor( int row, int col ) {
	Q3TableItem *curritem = item(row, col);
	if (curritem == NULL) {
		addUndoRecord(row, col, row, col);
	}
	Q3Table::setCellContentFromEditor(row, col);
}




void XmlTable::defaultSettings() {
	styleVector.setAutoDelete(TRUE);
	if (styleVector.isEmpty()) {
		for (int type = XmlTableItem::EMPTY; type <= XmlTableItem::COMMENT; type++) {
			styleVector.insert(type, new XmlTableStyle());
		}
	}
	QFont currFont;
	int align = Qt::AlignLeft | Qt::AlignTop;
	int wrap = Qt::TextWrapAnywhere;
	QString fg = "#000000", bg, name;
	for (int type = XmlTableItem::EMPTY; type <= XmlTableItem::COMMENT; type++) {
		XmlTableStyle* currStyle = styleVector[type];
		switch (type) {
			case XmlTableItem::ELEMENT:
				name = "element";
				currFont = QFont("Helvetica", 10, QFont::Bold, FALSE);
				align = Qt::AlignRight | Qt::AlignTop;
				wrap = Qt::TextSingleLine;
				bg = "#BBBBFF";
				break;
			case XmlTableItem::ATTR_KEY:
				name = "key";
				currFont = QFont("Arial", 9, QFont::Bold, FALSE);
				align = Qt::AlignHCenter | Qt::AlignBottom;
				wrap = Qt::TextSingleLine;
				bg = "#BBBBBB";
				break;
			case XmlTableItem::ATTR_VAL:
				name = "value";
				currFont = QFont("Arial", 10, QFont::Normal, FALSE);
				align = Qt::AlignHCenter | Qt::AlignTop;
				wrap = Qt::TextWrapAnywhere;
				bg = "#DDDDDD";
				break;
			case XmlTableItem::TEXT:
				name = "text";
				currFont = QFont("Andale Mono", 10, QFont::Normal, FALSE);
				align = Qt::AlignLeft | Qt::AlignTop;
				wrap = Qt::TextWordWrap;
				bg = "#FFBBBB";
				break;
			case XmlTableItem::CDATA:
				name = "CDATA";
				currFont = QFont("Andale Mono", 10, QFont::Normal, FALSE);
				align = Qt::AlignLeft | Qt::AlignTop;
				wrap = Qt::TextWrapAnywhere;
				bg = "#BBFFBB";
				break;
			case XmlTableItem::COMMENT:
				name = "comment";
				currFont = QFont("Times New Roman", 10, QFont::Normal, TRUE);
				align = Qt::AlignLeft | Qt::AlignTop;
				wrap = Qt::TextWordWrap;
				bg = "#FFFFFF";
				break;
			case XmlTableItem::EMPTY:
				name = "empty";
				currFont = QFont("Helvetica", 10, QFont::Normal, FALSE);
				align = Qt::AlignLeft | Qt::AlignTop;
				wrap = Qt::TextWordWrap;
				bg = "#FFFFFF";
				break;
		}
		currStyle->setName(name);
		currStyle->setFont(currFont.toString());
		currStyle->setAlignment(align);
		currStyle->setWordWrap(wrap);
		currStyle->setBgColor(bg);
		currStyle->setFgColor(fg);
	}
	minColumnWidth = 50;
}


void XmlTable::applySettings() {
	XmlTable::defaultSettings();
	QSettings settings;
	settings.setPath(TX_SETTINGS_PATH, "tx", QSettings::UserScope);
	settings.beginGroup("/tx/XmlTable");
	for (int type = XmlTableItem::EMPTY; type <= XmlTableItem::COMMENT; type++) {
		XmlTableStyle* currStyle = styleVector[type];
		QString name = currStyle->name();
		currStyle->setFont(
			settings.readEntry( QString("/%1/font").arg(name), currStyle->font().toString() )
			);
		currStyle->setAlignment(
			settings.readNumEntry( QString("/%1/align").arg(name), currStyle->alignment() )
			);
		currStyle->setWordWrap(
			settings.readNumEntry( QString("/%1/wrap").arg(name), currStyle->wordWrap() )
			);
		currStyle->setBgColor(
			settings.readEntry( QString("/%1/bgColor").arg(name), currStyle->bgColor().name() )
			);
		currStyle->setFgColor(
			settings.readEntry( QString("/%1/fgColor").arg(name), currStyle->fgColor().name() )
			);
	}
	settings.endGroup();
}


void XmlTable::saveSettings() {
	QSettings settings;
	settings.setPath(TX_SETTINGS_PATH, "tx", QSettings::UserScope);
	settings.beginGroup("/tx/XmlTable");
	for (int type = XmlTableItem::EMPTY; type <= XmlTableItem::COMMENT; type++) {
		XmlTableStyle* currStyle = styleVector[type];
		QString name = currStyle->name();
		settings.writeEntry( QString("/%1/font").arg(name), currStyle->font().toString() );
		settings.writeEntry( QString("/%1/align").arg(name), currStyle->alignment() );
		settings.writeEntry( QString("/%1/wrap").arg(name), currStyle->wordWrap() );
		settings.writeEntry( QString("/%1/bgColor").arg(name), currStyle->bgColor().name() );
		settings.writeEntry( QString("/%1/fgColor").arg(name), currStyle->fgColor().name() );
	}
	settings.endGroup();
}


XmlTableStyle* XmlTable::style(int type) {
	return styleVector[type];
}


QString XmlTable::dtdName() const {
	return(docTypeName);
}
